
class CSVAdapter(object):
    def __init__(self, db, **kwargs):
        self._connection_opts = kwargs
        self._connection = None
        self._db = db

    def select_all(self, mapping, obj_class):
        reader = self._db.reader(self.connection)
        for row in reader:
            obj = obj_class()
            yield mapping.row_to_object(row, obj)

    @property
    def connection(self):
        return self._connection if self._connection else self.connect()

    def connect(self):
        if not self._connection:
            self._connection = self._db.connect(**self._connection_opts)
        return self._connection

    def close_connection(self):
        if self._connection:
            self._connection.close()
            self._connection = None



from . import mapper


class QueryException(Exception):
    pass


class MultipleObjectsReturned(QueryException):
    pass


class ObjectNotFound(QueryException):
    pass


class GenericObject(object):
    pass


class Manager(object):
    def __init__(self, adapter, mapping, model=GenericObject):
        if isinstance(mapping, str):
            mapping = mapper.Mapping(mapping, scaffold=True)
        self._adapter = adapter
        self._mapping = mapping
        self.model = model

        if not mapping.columns() and mapping.scaffold:
            mapper.scaffold(adapter, mapping)

    @property
    def database(self):
        return self._adapter

    @property
    def mapping(self):
        return self._mapping

    def begin(self):
        self._adapter.start_transaction()

    def commit(self):
        self._adapter.commit_transaction()

    def rollback(self):
        self._adapter.rollback_transaction()

    def add(self, obj):
        self._adapter.insert(self._mapping, obj)

    def add_many(self, objs):
        self._adapter.insert_many(self._mapping, objs)

    def update(self, obj):
        self._adapter.update(self._mapping, obj)

    def update_many(self, objs):
        self._adapter.update_many(self._mapping, objs)

    def delete(self, obj):
        self._adapter.delete(self._mapping, obj)

    def elete_many(self, obj):
        self._adapter.delete_many(self._mapping, obj)

    def all(self):
        return self._adapter.select_all(self._mapping, self.model)

    def query(self, sql, *args):
        return self._adapter.raw(sql, self._mapping, self.model, args=args)



from __future__ import absolute_import

from psycopg2 import connect, paramstyle  # NOQA
from psycopg2.extras import DictCursor

import psycopg2 as dbapi  # NOQA

from ..adapters.genericsql import GenericSQLAdapter as DefaultAdapter  # NOQA


def initialize_connection(connection):
    connection.cursor_factory = DictCursor


def table_column_names(db, table, schema='public'):
    cur = db.cursor()
    cur.execute('''
        select column_name from information_schema.columns
        where table_schema = %(schema)s and table_name = %(table)s
        ''', {'table': table, 'schema': schema})
    return list(map(lambda x: x[0], cur.fetchall()))


def table_primary_key(db, table, schema='public'):
    cur = db.cursor()
    cur.execute('''
        SELECT column_name
        FROM information_schema.table_constraints
             JOIN information_schema.key_column_usage
             USING (constraint_catalog, constraint_schema, constraint_name,
                table_catalog, table_schema, table_name)
        WHERE constraint_type = 'PRIMARY KEY'
          AND (table_schema, table_name) = (%(schema)s, %(table)s)
        ORDER BY ordinal_position;
        ''', {'table': table, 'schema': schema})

    return list(map(lambda x: x[0], cur.fetchall()))



from __future__ import absolute_import

from sqlite3 import connect, Row, paramstyle  # NOQA
import sqlite3 as dbapi  # NOQA

from ..adapters.genericsql import GenericSQLAdapter as DefaultAdapter  # NOQA


def initialize_connection(connection):
    connection.isolation_level = 'DEFERRED'
    connection.row_factory = Row  # NOQA



from __future__ import absolute_import
from ..adapters.csv import CSVAdapter as DefaultAdapter  # NOQA

import csv


class Connection(object):
    def __init__(self, path, **opts):
        self.opts = opts
        self.path = path

    def close():
        pass


def connect(path, **opts):
    return Connection(path, **opts)


class DictResult(object):
    def __init__(self, reader):
        try:
            self._colnames = next(reader)
        except StopIteration:
            self._colnames = None
            self._rows = []
        else:
            self._rows = reader

    def __iter__(self):
        for row in self._rows:
            yield dict(zip(self._colnames, row))


def reader(connection):
    fh = open(connection.path, 'r')
    return DictResult(csv.reader(fh, *connection.opts))




class MappingError(Exception):
    pass


class ColumnNotDefined(MappingError):
    pass


class ColumnAlreadyMapped(MappingError):
    pass


class PropertyAlreadyMapped(MappingError):
    pass


class Mapping(object):
    def __init__(
            self, table, mapping=None, primary_key=None, scaffold=False):
        self._c2p = {}
        self._p2c = {}
        self._dc = []
        self._tablename = table
        self._pk = []
        self.scaffold = scaffold

        if mapping:
            for columnname, propname in mapping.items():
                self.add(columnname, propname)

        if primary_key:
            self.set_primary_key(primary_key)

    def add(self, column_name, property_name):
        if column_name in self._c2p:
            raise ColumnAlreadyMapped(column_name)

        if property_name in self._p2c:
            raise PropertyAlreadyMapped(column_name)

        self._c2p[column_name] = property_name
        self._p2c[property_name] = column_name

        if column_name not in self._pk:
            self._dc.append(column_name)

    def row_to_object(self, row, obj):
        for colname, propname in self._c2p.items():
            try:
                value = row[colname]
            except (IndexError, KeyError):
                value = None
            setattr(obj, propname, value)
        return obj

    def object_to_dict(self, obj):
        result = {}
        for propname, colname in self._p2c.items():
            try:
                result[colname] = getattr(obj, propname)
            except AttributeError:
                pass  # skip, won't be updated
        return result

    def columns(self):
        return self._c2p.keys()

    def data_columns(self):
        return self._dc[:]

    def primary_key(self):
        return self._pk[:]

    def set_primary_key(self, column_names):
        columns = self.columns()
        for column_name in column_names:
            if column_name not in columns:
                raise ColumnNotDefined(column_name)
        self._pk = column_names
        for column_name in self._pk:
            try:
                self._dc.remove(column_name)
            except ValueError:
                pass

    def properties(self):
        return self._p2c.keys()

    @property
    def db_table(self):
        return self._tablename


def scaffold(adapter, mapping):
    column_names = adapter._db.table_column_names(
            adapter, mapping.db_table)
    pk = adapter._db.table_primary_key(adapter, mapping.db_table)
    for name in column_names:
        mapping.add(name, name)
    mapping.set_primary_key(pk)



from .persistence import Manager  # NOQA
from .mapper import Mapping  # NOQA
from . import engines  # NOQA


def database(dbapi, adapter=None, **connection_opts):
    adapter = adapter or getattr(dbapi, 'DefaultAdapter')
    return adapter(dbapi, **connection_opts)





class Query(object):
    def __init__(self, manager):
        self._mapping = manager.mapping
        self._manager = manager


class Select(Query):
    def filter(self, filterspec):
        return self

    def __iter__(self):
        return self._manager.database.select(self._mapping, self._manager.model)



PARAM_STYLES = {
    'qmark': lambda x: '?',
    'numeric': lambda x: ':%d' % x[0],
    'named': lambda x: ':%s' % x[1],
    'format': lambda x: '%s',
    'pyformat': lambda x: '%%(%s)s' % x[1],
    }


def named_parameters(cols, data):
    return dict(map(lambda x: (x, data[x]), cols))


def sequential_parameters(cols, data):
    return map(lambda x: data[x], cols)


class GenericSQLAdapter(object):
    def __init__(self, db, **kwargs):
        self._connection_opts = kwargs
        self._connection = None
        self._db = db
        self.param = PARAM_STYLES[db.paramstyle]

        if db.paramstyle in ('named', 'pyformat'):
            self.create_parameters = named_parameters
        else:
            self.create_parameters = sequential_parameters

    def configure_connection(self, **params):
        if self.is_connected():
            raise RuntimeError('Connection is estabilished')
        self._connection_opts = params

    def cursor(self):
        return self.connection.cursor()

    def insert(self, mapping, obj):
        """Inserts the `obj` instance into database"""

        data = mapping.object_to_dict(obj)
        cols = mapping.data_columns()

        cur = self.cursor()

        sql = '''insert into %s (%s) values (%s)''' % (
            mapping.db_table,
            ','.join(cols),
            ','.join(map(self.param, enumerate(cols)))
        )

        cur.execute(sql, self.create_parameters(cols, data))

    def update(self, mapping, obj):
        """Updates the `obj` instance state in the database"""

        data = mapping.object_to_dict(obj)
        cols = mapping.data_columns()

        cur = self.cursor()

        all_columns = mapping.columns()

        set_params = map(
                lambda x: '%s=%s' % (x[1], self.param(x)), enumerate(cols))

        where_params = map(
                lambda x: '%s=%s' % (x[1], self.param(x)),
                enumerate(mapping.primary_key()))

        sql = '''update %s set %s where %s''' % (
            mapping.db_table,
            ','.join(set_params),
            ' and '.join(where_params),
            )

        cur.execute(sql, self.create_parameters(all_columns, data))

    def select_all(self, mapping, obj_class, limit=None, offset=None):
        cols = mapping.columns()
        sql = 'select %s from %s'

        if offset:
            sql += ' offset %d' % offset

        if limit:
            sql += ' limit %d' % limit

        sql = sql % (','.join(cols), mapping.db_table)

        return self.raw(sql, mapping, obj_class)

    def raw(self, sql, mapping, obj_class, args=None):
        """
        Instantiate objects of `obj_class` using `mapping`
        from raw `sql`
        """

        args = args or []
        cur = self.cursor()
        ctx = {
                'table': mapping.db_table,
                }
        sql = sql.replace('%', '%%')
        cur.execute(sql % ctx, args)

        res = cur.fetchall()

        for row in res:
            obj = obj_class()
            yield mapping.row_to_object(row, obj)

    def connect(self):
        self._connection = self._db.connect(**self._connection_opts)
        try:
            self._db.initialize_connection(self._connection)
        except AttributeError:
            pass
        return self._connection

    def is_connected(self):
        return self._connection is not None

    def close_connection(self):
        if self.is_connected():
            self._connection.close()
            self._connection = None

    @property
    def connection(self):
        return self._connection if self.is_connected() else self.connect()

    def init(self, **connection_opts):
        self._connection_opts = connection_opts

    def start_transaction(self):
        cur = self.cursor()
        cur.execute('begin')

    def commit_transaction(self):
        self.connection.commit()

    def rollback_transaction(self):
        self.connection.rollback()



from setuptools import setup, find_packages

setup(name='worm-dao',
      version='0.1.4',
      description='Worst or wonderful DAO library',
      classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Topic :: Database :: Front-Ends",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
        ],
      author='Marcin Nowak',
      author_email='marcin.j.nowak@gmail.com',
      url='https://github.com/marcinn/worm',
      keywords='dao orm database',
      packages=find_packages('.'),
      include_package_data=True,
      zip_safe=True,
      )

